# $NetBSD: t_cd9660_regression.sh,v 1.0 2014/06/12 18:00:00 me Exp $
#
# Copyright (c) 2014 The NetBSD Foundation, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#
# This script tests the cd9660 filesystem driver for ISO 9660 with Rock Ridge.
# The particular test cases are:
#
#    pr_kern_48787   Verifies that ISO 9660 Directory Records above 4 GiB
#                    lead to the expected very high inode numbers.
#
#    rock_ridge_rgr  Checks whether Rock Ridge file types are interpreted
#                    properly.
#
#    mount_s         Verifies the effect of mount_cd9660 option -s.
#
#    large_file      Checks recognition and readability of a large file.
#
#    exotic          Checks the handling of exotic ISO 9660 features:
#                    - File "01" exists in three versions.
#                    - File "07" has an undigestible first file section.
#                    - File "09" contains an ISO 9660 Extended Attribute block.
#                    - File "12" has an associated file (Rock Ridge name "11").
# 
# Two test images are provided:
#
#    cd9660_rgr.imageat0.bz2.uue
#    cd9660_rgr.image.at2114008.bz2.uue
#                    Construction kit for an ISO 9660 image of 4,329,758,720
#                    bytes. On suitable filesystems it occupies only a few
#                    hundred KB.
#  
#    exoten.iso.bz2.uue
#                    Unpacks to a small ISO 9660 image of 244 KiB.
#
#


# The script is equipped with some optional code which provides substitutes
# for the used features of atf and rump.
#
# Set this to 0 if this script shall be run under atf
atf_mockup=1


# Global setup:

# Commands to be used for mounting and unmountig of the ISO 9660 test images
rump_cd9660="rump_cd9660"
umount="umount"

# Leaf names of the test images
rgr_image="cd9660_rgr.image"
exotic_image="exoten.iso"

# Path to not yet existing directory for mounting the test images
mntpnt="$(pwd)/mnt"


##### Start of atf mock-up code #######################################
#
# This code allows to test the functions without having runnable atf
#
if [ "$atf_mockup" = 1 ]; then

#	rump_cd9660="own_rump_cd9660"
# 	umount="own_umount"

	devno="0"
	rump_dev="/dev/vnd${devno}d"
	dev_vnconfiged=0

	failure_count=0

	atf_set() {
		if [ "$1" = "descr" ]; then
			echo >&2
			echo "Starting test case which $2" >&2
			echo >&2
		fi
	}

	atf_fail() {
		echo "FAILURE: " "$@" >&2
		failure_count=$(expr $failure_count + 1)
	}

	atf_add_test_case() {
		"$1"_head
		"$1"_body
		"$1"_cleanup
	}

	atf_get_srcdir() {
		pwd
	}

	atf_check() {
		pattern=$(echo "$2" | sed -e 's/match://')
		shift 2
		text="$("$@")"
		if echo "$text" | grep "$pattern" >/dev/null ; then
			dummy=dummy
		else
			atf_fail "Pattern mismatch with command: ""$@"
			echo "Pattern: ${pattern}" >&2
			echo "Output : '${text}'" >&2
		fi
	}

	own_rump_cd9660() {
		# Get last two arguments
		rump_opts=""
		rump_img=""
		rump_mnt=""
		for i in "$@"
		do
			if [ -n "$rump_img" ] ; then
				rump_opts="${rump_opts} ${rump_img}"
			fi
			rump_img="$rump_mnt"
			rump_mnt="$i"
		done
		if [ -n "$rump_img" ] ; then
			vnconfig -c vnd"$devno" "$rump_img" && dev_vnconfiged=1
		fi
		mount_cd9660 $rump_opts "$rump_dev" "$rump_mnt"
	}

	own_umount() {
		umount "$@"
		if [ "$dev_vnconfiged" = 1 ]; then
			vnconfig -u vnd"$devno"
			dev_vnconfiged=0
		fi
	}

	rm ${rgr_image}.at0.bz2 ${rgr_image}.at2114008.bz2 ${exotic_image}.bz2
	uudecode -p ${rgr_image}.at0.bz2.uue > ${rgr_image}.at0.bz2
	uudecode -p ${rgr_image}.at2114008.bz2.uue > ${rgr_image}.at2114008.bz2
	uudecode -p ${exotic_image}.bz2.uue > ${exotic_image}.bz2

fi
##### End of atf mock-up code #########################################


cd9660_create_rgr_image() {
	bunzip2 < $(atf_get_srcdir)/${rgr_image}.at0.bz2 > ${rgr_image}
	bunzip2 < $(atf_get_srcdir)/${rgr_image}.at2114008.bz2 \
		| dd bs=2048 seek=2114008 of=${rgr_image}
}

cd9660_cleanup_common() {
	if [ ! -f done ]; then
		if [ "x${mntpnt}" != "x" ]; then
			"$umount" -f ${mntpnt} || true
		fi
	else
		rm "done"
	fi
	if [ -d "$mntpnt" ]; then
		rmdir "$mntpnt"
	fi
}


pr_kern_48787_head() {
	atf_set "descr" "Verifies 32bit overflow isssues from PR kern/48787 are fixed"
	atf_set "require.user" "root"
	atf_set "require.progs" "rump_cd9660 bunzip2 dd stat"
	atf_set "timeout" 6000
	atf_set "use.fs" "true"
}

pr_kern_48787_body() {
	cd9660_create_rgr_image
	mkdir ${mntpnt}
	"$rump_cd9660" -o norrip $(atf_get_srcdir)/${rgr_image} ${mntpnt}
	if [ ! -r ${mntpnt}/small_file ]; then
		atf_fail "${mntpnt}/small_file does not exist"
	fi
	if [ ! -r ${mntpnt}/my/large_file ]; then
		atf_fail "${mntpnt}/my/large_file does not exist"
	fi
	"$umount" ${mntpnt}
	"$rump_cd9660" $(atf_get_srcdir)/${rgr_image} ${mntpnt}
	if [ ! -r ${mntpnt}/small_file ]; then
		atf_fail "${mntpnt}/small_file does not exist"
	fi
	if [ ! -r ${mntpnt}/my/large_file ]; then
		atf_fail "${mntpnt}/my/large_file does not exist"
	fi
	echo "this assumes current cd9660 inode encoding - adapt on changes"
	atf_check -o match:"^4329607714$" stat -f "%i" ${mntpnt}/small_file
	atf_check -o match:"^281479306324160$" stat -f "%i" ${mntpnt}/my/large_file
	"$umount" ${mntpnt}
	touch "done"
}

pr_kern_48787_cleanup() {
	cd9660_cleanup_common
}


rock_ridge_rgr_head() {
	atf_set "descr" "Checks Rock Ridge file types for regressions"
	atf_set "require.user" "root"
	atf_set "require.progs" "rump_cd9660 bunzip2 dd md5"
	atf_set "timeout" 6000
	atf_set "use.fs" "true"
}

rock_ridge_rgr_body() {
	cd9660_create_rgr_image
	mntpnt=$(pwd)/mnt
	mkdir ${mntpnt}
	"$rump_cd9660"  $(atf_get_srcdir)/${rgr_image} ${mntpnt}

	regfile_md5="b346da73f1de37861749447106a6e745"
	path="${mntpnt}/reg/tar"
	md5=$(md5 <"$path")
	if [ "$md5" != "$regfile_md5" ]; then
		atf_fail "MD5 mismatch with content of '${path}'"
	fi

	dirs_dev_reg="${mntpnt}/dev/test.fifo ${mntpnt}/dev/wd1e ${mntpnt}/reg/tar ${mntpnt}/reg/to_regfile"
	dev_reg="$(echo "$mntpnt"/dev/* "$mntpnt"/reg/*)"
	if [ "$dev_reg" != "$dirs_dev_reg" ]; then
		atf_fail "Wrong content reported for: ${mntpnt}/dev/* ${mntpnt}/reg/*"
	fi

	path="${mntpnt}/dev"
	blockdev="wd1e"
	dev_lsl="br--------  1 1000  5  0, 12 May 14 14:33 wd1e"
	lsl="$(ls -ln "$path" | grep "$blockdev")"
	if [ "$lsl" != "$dev_lsl" ]; then
		atf_fail "Wrong output from ls -ln ${path}"
	fi
	
	path="${mntpnt}/dev/test.fifo"
	to_fifo="Through the fifo"
	echo "$to_fifo" >"$path" &
	from_fifo=$(cat "$path")
	if [ "$to_fifo" != "$from_fifo" ]; then
		atf_fail "Failure to send text through FIFO ${path}"
	fi

	symlink="to_regfile"
	path="${mntpnt}/reg/to_regfile"
	link="$(readlink "$path")"
	if [ "$link" != "tar" ]; then
		atf_fail "Wrong reply from readlink ${path}: ${link} != 'tar'"
	fi

	"$umount" ${mntpnt}
	touch "done"
}

rock_ridge_rgr_cleanup() {
	cd9660_cleanup_common
}


mount_s_head() {
	atf_set "descr" "Verifies implementation of mount_cd9660 option -s"
	atf_set "require.user" "root"
	atf_set "require.progs" "rump_cd9660 bunzip2 dd"
	atf_set "timeout" 6000
	atf_set "use.fs" "true"
}

mount_s_body() {
	cd9660_create_rgr_image
	mntpnt=$(pwd)/mnt
	mkdir ${mntpnt}
	"$rump_cd9660" -s 32 $(atf_get_srcdir)/${rgr_image} ${mntpnt}

	path="${mntpnt}"
	mnts_lsl='drwxr-xr-x  1 1000  1000  2048 May  6 15:30 my'
	lsl="$(ls -ln "$path" | tail +2)"
	if [ "$lsl" != "$mnts_lsl" ]; then
		atf_fail "Wrong output from -s 32 ; ls -ln ${path}"
	fi

	"$umount" ${mntpnt}
	"$rump_cd9660" -s 2114016 $(atf_get_srcdir)/${rgr_image} ${mntpnt}

	path="${mntpnt}"
	mnts_lsl='drwxr-xr-x  1 1000  1000  2048 May  6 15:30 my
-rw-r--r--  1 1000  1000     6 May  6 15:34 small_file'
	lsl="$(ls -ln "$path" | tail +2)"
	if [ "$lsl" != "$mnts_lsl" ]; then
		atf_fail "Wrong output from -s 2114016 ; ls -ln ${path}"
	fi

	"$umount" ${mntpnt}
	"$rump_cd9660" -s 0 $(atf_get_srcdir)/${rgr_image} ${mntpnt}

	path="${mntpnt}"
	mnts_lsl='dr-x------  1 1000  0     2048 May  3 14:58 dev
dr-x------  1 1000  1000  2048 May  6 15:30 my
dr-x------  1 1000  0     2048 Jan 19 14:41 reg
-r--------  1 1000  1000     6 May  6 15:34 small_file'
	lsl="$(ls -ln "$path" | tail +2)"
	if [ "$lsl" != "$mnts_lsl" ]; then
		atf_fail "Wrong output from -s 0 ; ls -ln ${path}"
	fi

	"$umount" ${mntpnt}
	touch "done"
}

mount_s_cleanup() {
	cd9660_cleanup_common
}


large_file_head() {
	atf_set "descr" "Checks handling of file >= 4 GiB"
	atf_set "require.user" "root"
	atf_set "require.progs" "rump_cd9660 bunzip2 dd time md5"
	atf_set "timeout" 6000
	atf_set "use.fs" "true"
}

large_file_body() {
	cd9660_create_rgr_image
	mntpnt=$(pwd)/mnt
	mkdir ${mntpnt}
	"$rump_cd9660"  $(atf_get_srcdir)/${rgr_image} ${mntpnt}

	path="${mntpnt}/my"
	large_lsl="-r--------  1 1000  1000  4329375744 May  6 15:30 large_file"
	lsl="$(ls -ln "$path" | grep "large_file")"
	if [ "$lsl" != "$large_lsl" ]; then
		atf_fail "Wrong output from ls -ln ${path}"
	fi

	path="${mntpnt}/my/large_file"

	echo "Reading first 32 kB of large file content ..." >&2
	first_md5="1f5dd6ce1d849969be08fd5ce32561c3"
	md5=$(dd if="$path" bs=2048 count=16 | md5)
	if [ "$md5" != "$first_md5" ]; then
		atf_fail "MD5 mismatch with first 32 kB of '${path}'"
	fi

	echo "Reading last 32 kB of large file content ..." >&2
	last_md5="a8e15a4c2f03eef984ed629108e0edb5"
	md5=$(dd if="$path" bs=2048 skip=2113937 | md5)
	if [ "$md5" != "$last_md5" ]; then
		atf_fail "MD5 mismatch with last 32 kB of '${path}'"
	fi

	"$umount" ${mntpnt}
	touch "done"
}

large_file_cleanup() {
	cd9660_cleanup_common
}


exotic_head() {
	atf_set "descr" "Checks handling of exotic ISO 9660 features"
	atf_set "require.user" "root"
	atf_set "require.progs" "rump_cd9660 bunzip2 dd"
	atf_set "timeout" 6000
	atf_set "use.fs" "true"
}

exotic_body() {
	bunzip2 < "${exotic_image}.bz2" > "$exotic_image"
	mntpnt=$(pwd)/mnt
	mkdir ${mntpnt}
	"$rump_cd9660"  $(atf_get_srcdir)/${exotic_image} ${mntpnt}

	echo \
      "Note: Error messages about unaligned non-last file section are expected"

	path="${mntpnt}/exotic"
	exo_lsl='-rw-r--r--  1 1000  1000  4096 May 26 08:00 00
-rw-r--r--  1 1000  1000  4096 May 26 08:00 01
-rw-r--r--  1 1000  1000  4096 May 26 08:00 02
-rw-r--r--  1 1000  1000  8192 May 26 08:00 03
-rw-r--r--  1 1000  1000  4096 May 26 08:00 05
-rw-r--r--  1 1000  1000  4096 May 26 08:00 08
-rw-r--r--  1 1000  1000  2048 May 26 08:00 09
-rw-r--r--  1 1000  1000  4096 May 26 08:00 10
-rw-r--r--  1 1000  1000  4096 May 26 08:00 11
-rw-r--r--  1 1000  1000  4096 May 26 08:00 12
-rw-r--r--  1 1000  1000  4096 May 26 08:00 13'
	lsl="$(ls -ln "$path" 2>/dev/null | head -12 | tail -11)"
	if [ "$lsl" != "$exo_lsl" ]; then
		atf_fail "Wrong output from -o rrip ; ls -ln ${path}"
	fi

	"$umount" ${mntpnt}
	"$rump_cd9660" -o norrip,gens \
			$(atf_get_srcdir)/${exotic_image} ${mntpnt}

	exo_lsl='-r-xr-xr-x  1 0  0  4096 May 26 08:00 00.;1
-r-xr-xr-x  1 0  0  4096 May 26 08:00 01.;1
-r-xr-xr-x  1 0  0  4096 May 26 08:00 01.;2
-r-xr-xr-x  1 0  0  8192 May 26 08:00 01.;3
-r-xr-xr-x  1 0  0  4096 May 26 08:00 05.;1
-r-xr-xr-x  1 0  0  4096 May 26 08:00 08.;1
-r-xr-xr-x  1 0  0  2048 May 26 08:00 09.;1
-r-xr-xr-x  1 0  0  4096 May 26 08:00 10.;1
-r-xr-xr-x  1 0  0  4096 May 26 08:00 12.;1
-r-xr-xr-x  1 0  0  4096 May 26 08:00 13.;1'
	lsl="$(ls -ln "$path" 2>/dev/null | head -11 | tail -10)"
	if [ "$lsl" != "$exo_lsl" ]; then
		atf_fail "Wrong output from -o norrip,gens ; ls -ln ${path}"
	fi
	exo_lsl='-r-xr-xr-x  1 0  0  4096 May 26 08:00 =12.;1'
	lsl="$(ls -ln "$path" 2>/dev/null | tail -1)"
	if [ "$lsl" != "$exo_lsl" ]; then
		atf_fail "Wrong output from -o norrip,gens ; ls -ln ${path} | tail -1"
	fi

	"$umount" ${mntpnt}
	"$rump_cd9660" -o norrip $(atf_get_srcdir)/${exotic_image} ${mntpnt}

	exo_lsl='-r-xr-xr-x  1 0  0  4096 May 26 08:00 00
-r-xr-xr-x  1 0  0  8192 May 26 08:00 01
-r-xr-xr-x  1 0  0  4096 May 26 08:00 05
-r-xr-xr-x  1 0  0  4096 May 26 08:00 08
-r-xr-xr-x  1 0  0  2048 May 26 08:00 09
-r-xr-xr-x  1 0  0  4096 May 26 08:00 10
-r-xr-xr-x  1 0  0  4096 May 26 08:00 12
-r-xr-xr-x  1 0  0  4096 May 26 08:00 13'
	lsl="$(ls -ln "$path" 2>/dev/null | head -9 | tail -8)"
	if [ "$lsl" != "$exo_lsl" ]; then
		atf_fail "Wrong output from -o norrip ; ls -ln ${path}"
	fi
	exo_lsl='-r-xr-xr-x  1 0  0  4096 May 26 08:00 =12'
	lsl="$(ls -ln "$path" 2>/dev/null | tail -1)"
	if [ "$lsl" != "$exo_lsl" ]; then
		atf_fail "Wrong output from -o norrip ; ls -ln ${path} | tail -1"
	fi

	"$umount" ${mntpnt}
	"$rump_cd9660" -o norrip,extatt \
			$(atf_get_srcdir)/${exotic_image} ${mntpnt}
	exo_lsl="-r-x--xr--  1 1234  5678  2048 Jun  4 07:51 ${path}/09"
	lsl="$(ls -ln "${path}/09")"
	if [ "$lsl" != "$exo_lsl" ]; then
		atf_fail "Wrong output from -o norrip,extatt ; ls -ln ${path}/09 : '${lsl}'"
	fi

	"$umount" ${mntpnt}
	touch "done"
}

exotic_cleanup() {
	cd9660_cleanup_common
}


atf_init_test_cases() {

	atf_add_test_case pr_kern_48787

	atf_add_test_case rock_ridge_rgr

	atf_add_test_case mount_s

	atf_add_test_case large_file

	atf_add_test_case exotic
}

##### Start of atf mock-up code #######################################
#
# This code runs the test cases and cleans up aftwerwards.
#
if [ "$atf_mockup" = 1 ]; then

	atf_init_test_cases

	if [ -d "$mntpnt" ] ; then
		# Should only happen if previous test was aborted
		"$umount" "$mntpnt"
		rmdir "$mntpnt"
	fi
	if [ -f "$rgr_image" ] ; then
		rm "$rgr_image"
	fi
	if [ -f "$exotic_image" ] ; then
		rm "$exotic_image"
	fi

	echo
	echo "========================================================"
	if [ "$failure_count" = 0 ]; then
		echo "All is well."
	else
		echo "FAILURES: ${failure_count}"
	fi
	echo

fi
##### End of atf mock-up code #########################################

